<!--
The request to this GraphQL server provided the header "Accept: text/html"
and as a result has been presented GraphiQL - an in-browser IDE for
exploring GraphQL.
If you wish to receive JSON, provide the header "Accept: application/json" or
add "&raw" to the end of the URL within a browser.
-->
{% load static %}
<!DOCTYPE html>
<html>
<head>
  <style>
    html, body, #editor {
      height: 100%;
      margin: 0;
      overflow: hidden;
      width: 100%;
    }
  </style>
  <link href="https://cdn.jsdelivr.net/npm/graphiql@{{graphiql_version}}/graphiql.min.css"
        integrity="{{graphiql_css_sri}}"
        rel="stylesheet"
        crossorigin="anonymous" />
  <link href="https://cdn.jsdelivr.net/npm/@graphiql/plugin-explorer@{{graphiql_plugin_explorer_version}}/dist/style.css"
        integrity="{{graphiql_plugin_explorer_css_sri}}"
        rel="stylesheet"
        crossorigin="anonymous" />
  <script src="https://cdn.jsdelivr.net/npm/whatwg-fetch@{{whatwg_fetch_version}}/dist/fetch.umd.js"
          integrity="{{whatwg_fetch_sri}}"
          crossorigin="anonymous"></script>
  <script src="https://cdn.jsdelivr.net/npm/react@{{react_version}}/umd/react.production.min.js"
          integrity="{{react_sri}}"
          crossorigin="anonymous"></script>
  <script src="https://cdn.jsdelivr.net/npm/react-dom@{{react_version}}/umd/react-dom.production.min.js"
          integrity="{{react_dom_sri}}"
          crossorigin="anonymous"></script>
  <script src="https://cdn.jsdelivr.net/npm/graphiql@{{graphiql_version}}/graphiql.min.js"
          integrity="{{graphiql_sri}}"
          crossorigin="anonymous"></script>
  <script src="https://cdn.jsdelivr.net/npm/graphql-ws@{{subscriptions_transport_ws_version}}/umd/graphql-ws.min.js"
          integrity="{{subscriptions_transport_ws_sri}}"
          crossorigin="anonymous"></script>
  <script src="https://cdn.jsdelivr.net/npm/@graphiql/plugin-explorer@{{graphiql_plugin_explorer_version}}/dist/graphiql-plugin-explorer.umd.js"
          integrity="{{graphiql_plugin_explorer_sri}}"
          crossorigin="anonymous"></script>
</head>
<body>
  <div id="editor"></div>
  {% csrf_token %}
  <script type="application/javascript">
    window.GRAPHENE_SETTINGS = {
    {% if subscription_path %}
      subscriptionPath: "{{subscription_path}}",
    {% endif %}
      graphiqlHeaderEditorEnabled: {{ graphiql_header_editor_enabled|yesno:"true,false" }},
      graphiqlShouldPersistHeaders: {{ graphiql_should_persist_headers|yesno:"true,false" }},
    };
  </script>
  <script src="{% static 'graphene_django/graphiql.js' %}"></script>
</body>
</html>
